/*
 * Copyright (c) 2022 VMware, Inc. All Rights Reserved.
 */

const {cspFFClient, FFEvents} = require('./feature_flag');
const {reqCtxt} = require('./request_context');
const {mapSchema, getDirective, MapperKind} = require('@graphql-tools/utils');
const {smDeployment, ffConfig} = require('./config');
const {log} = require("./logger");

//This currently only filters fieldDefs, input object types, arguments.
//As our use cases increase we can add other items to the list like arg def, enum values etc.
module.exports.featureFlagDirectiveTransformer =  function featureFlagDirectiveTransformer(schema, directiveName='featureFlag') {
    const evalFf = (obj) => {
        const directives = getDirective(schema, obj, directiveName) || [];
        const names = directives.flatMap(directive => directive.name);
        return names.every(evaluateFF) ? obj : null;
    }
    return mapSchema(schema, {
        // Executes once for each object field definition in the schema
        [MapperKind.ARGUMENT]: evalFf,
        [MapperKind.INPUT_OBJECT_TYPE]: evalFf,
        [MapperKind.INPUT_OBJECT_FIELD]: evalFf,
        [MapperKind.OBJECT_FIELD]: evalFf
    });
}
//Collecting the all feature flag name in the schema to avoid iteration through the entire schema for each orgs schema filter.
module.exports.extractSchemaFeatureFlagNames = function extractSchemaFeatureFlagNames(schema, directiveName = 'featureFlag') {
    const featureFlagsInSchema = [];
    const addFfs = (obj) => {
        const directives = getDirective(schema, obj, directiveName) || [];
        const names = directives.flatMap(directive => directive.name);
        featureFlagsInSchema.push(...names);
        return obj;
    };
    mapSchema(schema, {
        // Executes once for each object field definition in the schema
        [MapperKind.ARGUMENT]: addFfs,
        [MapperKind.INPUT_OBJECT_TYPE]: addFfs,
        [MapperKind.INPUT_OBJECT_FIELD]: addFfs,
        [MapperKind.OBJECT_FIELD]: addFfs
    });
    featureFlagsInSchema.sort();
    return new Set(featureFlagsInSchema);
}

const evaluateFF = function evaluateFF(ffName) {
    let ctxt = {
        userAcct: reqCtxt.get('user'),
        orgId: reqCtxt.get('orgId')
    };

    // read FF Json configuration, which will only exist in sm mode
    const ffJson = ffConfig()
    if (ffJson) {
        const obj =  ffJson.FeatureFlags.find(o => o.name === ffName);
        let result = false;
        if (obj && obj.value === true) {
            result = true;
        }
        log.info(`FF: ${ffName} = ${result}`)
        return result;
    }

    //This method smartly figures out the evaluation based on data passed, for unauth users, if the ctxt
    //has no userid and orgid this will evaluate to true only if the value is globally set to true.
    return cspFFClient.evaluate(ffName, ctxt); //boolean CSP FF value
}

module.exports.cspFFClient = cspFFClient;
module.exports.FFEvents = FFEvents;
module.exports.evaluateFF = evaluateFF;
