/*
 * Copyright (c) 2022 VMware, Inc. All Rights Reserved.
 */

const {print} = require('graphql');
const {extend} = require('got');
const {log} = require('./logger');
const {createClient} = require('graphql-sse');
const {reqCtxt} = require("./request_context");
// the inbuilt fetch and abort controller doesnt seem to be playing nice with graphql-sse.
// This causes the connection to be still open after the subscription client has stopped. Hence we are using
// the below libs to ensure connection is closed.
const { AbortController } = require('node-abort-controller'); 
const {fetch} = require('@whatwg-node/fetch');
const {gatewayConfig} = require('./config')

const SUPPORTED_FILE_CONTENT_TYPES = gatewayConfig().supportedFileContentTypes;

//Common extensions to got.
const client = extend({
    retry: {
        limit: 0
    },
    headers: {
        'Content-Type': 'application/json'
    },
    responseType: 'json',
    hooks: {
        beforeError: [
            handleError
        ],
        afterResponse: [
            (response) => {
                // log.debug(`HTTP ${response.req.method} ${response.url} ${response.statusCode} ${response.timings.phases.total}ms`);
                return response;
            }
        ]
    }
});

// Builds a remote schema executor function,
// customize any way that you need (auth, headers, etc).
// Expects to receive an object with "document" and "variable" params,
// and asynchronously returns a JSON response from the remote.
module.exports = function makeRemoteExecutor(url, subscriptionUrl, {timeout = 500} = {}) {
    const clientInstance = client.extend({
        url: url,
        timeout: {
            request: timeout
        }
    });

    const subscriptionClient = createClient({
        url: subscriptionUrl,
        fetchFn: fetch,
        abortControllerImpl: AbortController,
        retryAttempts: 0,
        headers: () => {
            let {headers} =  reqCtxt.get("subscriptionCtxt"); //getting headers from cls-hooked for auth etc.
            return headers;
        }
    });

    return async ({document, context, variables, operationName, operationType}) => {
        const query = typeof document === 'string' ? document : print(document);
        if (variables) {
            for (const [key, value] of Object.entries(variables)) {
                const mimetype = variables[key]?.type;
                // Currently, we support and have tested for ONLY application/octet-stream and application/zip mime-types.
                // If any other mime-types needs to be supported we would require testing and might also require custom handling of the code.
                if (SUPPORTED_FILE_CONTENT_TYPES?.indexOf(mimetype) > -1) {
                    log.debug("Mime Type:", mimetype);
                    let fileObject = {
                        originalname: variables[key]?.name,
                        mimetype,
                        buffer: Buffer.from(await variables[key]?.arrayBuffer()),
                        size: variables[key]?.size,
                    };
                    variables[key] = fileObject;
                }
            }
        }
        if(operationType !== 'subscription'){
            // logs to each microservice including all variables are too verbose so disable
            // if (log.isDebugEnabled()) {
            //     log.debug(`Request to ${url}: ${query} ` + JSON.stringify(variables));
            // }

            let {body} = await clientInstance.post({
                headers: context?.headers,
                json: {query, variables}
            });
            return body;
        } else {

            const request = {
                query: print(document),
                variables,
                operationName
            }

            let deferred = null;
            const pending = [];
            let throwMe = null,
                done = false;
            const dispose = subscriptionClient.subscribe(request, {
                next: (data) => {
                    log.debug(`Subscription Data rcvd: ${JSON.stringify(data)}`);
                    pending.push(data);
                    deferred?.resolve(false);
                },
                error: (err) => {
                    log.debug('Subscription Error rcvd: ',err);
                     if (err?.response?.status === 400 || err?.response?.status  === 401){
                         pending.push({"errors": [ {"message": "UNAUTHENTICATED"}]});
                     } else {
                         pending.push({"errors": [{"message": "Error occurred while processing this subscription."}]});
                     }
                      done=true;
                      deferred.resolve(false);
                },
                //Currently DGS does not send this event. So the client side (this class and the graphql-sse client lib) does not understand that the subscription is complete and will 
                //retry on a closure of connection from DGS. This is currently expected. I have raised a bug with the DGS team, 
                //https://github.com/Netflix/dgs-framework/issues/1435 
                //For our current subscription use cases this is not a show stopper, but in case a subscription provider has a limited number of events being sent 
                //say action completed, the client (UI) that called stitching will have to close the connection based on the terminal state until DGS fixes the bug.
                //When DGS fixes the bug we have no changes that need to be done here.
                complete: () => { 
                    done = true;
                    deferred?.resolve(true);
                },
            });
            return {
                [Symbol.asyncIterator]() {
                    return this;
                },
                async next() {
                    if (pending.length) return {value: pending.shift()};
                    if (done) return {done: true, value: undefined};
                    if (throwMe) throw throwMe;
                    return (await new Promise((resolve, reject) => (
                            deferred = {resolve, reject}))
                    )
                        ? {done: true, value: undefined}
                        : {value: pending.shift()};
                },
                async throw(err) {
                    throw err;
                },
                async return() {
                    dispose();
                    return {done: true, value: undefined};
                }
            };

        }
    };
};

function handleError(error) {
    let {response = {}} = error; //sometimes we may not have a response; ex: in case of a timeout.
    level = "error"
    if (error.code === "ECONNREFUSED") {
        level = "debug"
        error.message = "Connection refused"
    } else if (error.name === "TimeoutError") {
        error.message = "Timed out while executing request." //change message to hide timeout details.
    } else if (response.statusCode === 401) {
        error.message = "UNAUTHENTICATED";
    } else if (response.statusCode === 403) {
        error.message = "PERMISSION_DENIED";
    } else {
        //default handler change detailed error message to just status if available, if not send generic message
        error.message = response.statusMessage || "Error occurred while processing request";
    }
    log.log(level, `Request error - code: '${error.code}' message:'${error.message}' url:'${error.request.requestUrl}'  body:'${JSON.stringify(error.options.json)}' response:'${JSON.stringify(response.body)}' time:'${error.timings.phases.total}ms'`)
    return error;
}

