/*
 * Copyright (c) 2022 VMware, Inc. All Rights Reserved.
 */

const {gatewayConfig} = require('./config');
const {pruneSchema, filterSchema} = require("@graphql-tools/utils");
const {
    featureFlagDirectiveTransformer,
    extractSchemaFeatureFlagNames,
    evaluateFF,
    cspFFClient,
    FFEvents
} = require("./feature_flag_directive_schema_transformer");
const {featureStabilityDirectiveTransformer} = require("./feature_stability_directive_schema_transformer");
const {internalDirectiveTransformer} = require("./internal_directive_schema_transformer");
const {reqCtxt} = require('./request_context');
const {LRUCache} = require('lru-cache')
const {log} = require("./logger");
// Cache TTL kept intentionally infinite. Cache gets cleared on schema changes and any FF changes
const hashToSchemaCache = new LRUCache({max: 150, ttl: 0});
const orgFeatureFlagValuesHashCache = new LRUCache({max: gatewayConfig().orgSchemaCacheSize, ttl: 0});
const objectHash = require('object-hash');
const schemaFeatureFlagsSet = new Set();
const PRIVATE = "private";
const PUBLIC_SCHEMA = "PUBLIC";
const PUBLIC_SCHEMA_WITHOUT_FF_FILTER = "PUBLIC_SCHEMA_WITHOUT_FF_FILTER"

function createFilteredPublicGraphQLSchema(graphQLSchema) {
    let org = reqCtxt.get("orgId") || PUBLIC_SCHEMA;
    if (!orgFeatureFlagValuesHashCache.has(org)) {
        let startTime = Date.now();
        if (!hashToSchemaCache.has(PUBLIC_SCHEMA_WITHOUT_FF_FILTER)) {
            graphQLSchema = filterUnderScoreElementsWithOutPruning(graphQLSchema);
            let underScoreFilterFinishedTime = Date.now();
            log.debug(`filtered the underscore elements in the public schema ${underScoreFilterFinishedTime - startTime}ms`);

            graphQLSchema = internalDirectiveTransformer(graphQLSchema);
            let internalDirectiveFilterFinishedTime = Date.now();
            log.debug(`filtered the internal directive elements in the public schema ${internalDirectiveFilterFinishedTime - underScoreFilterFinishedTime}ms`);

            graphQLSchema = featureStabilityDirectiveTransformer(graphQLSchema);
            let featureStabilityDirectiveFilterFinishedTime = Date.now();
            log.debug(`filtered the feature stability directive elements in the public schema ${featureStabilityDirectiveFilterFinishedTime - internalDirectiveFilterFinishedTime} ms`);

            //adding the PUBLIC schema directly into schema cache due to more frequent usage for cacahe missed orgs requests.
            hashToSchemaCache.set(PUBLIC_SCHEMA_WITHOUT_FF_FILTER, graphQLSchema);
            orgFeatureFlagValuesHashCache.set(PUBLIC_SCHEMA_WITHOUT_FF_FILTER, PUBLIC_SCHEMA_WITHOUT_FF_FILTER);
            let publicSchemaFilterFinishedTime = Date.now();
            log.debug(`added common schema into cache in ${publicSchemaFilterFinishedTime - startTime}ms. Total ${hashToSchemaCache.size} entries in the cache.`);

            let ffNames = extractSchemaFeatureFlagNames(graphQLSchema);
            log.debug(`schema feature flag names extracted in ${Date.now() - publicSchemaFilterFinishedTime}ms.`);
            subscribeFeatureFlagEvents(ffNames);
            schemaFeatureFlagsSet.clear();
            for (ffname of ffNames) {
                schemaFeatureFlagsSet.add(ffname);
            }
        }
        let orgFeatureFlagValuesHash = computeOrgFeatureFlagValuesHash(org);
        let orgFeatureFlagvaluesHashFinisedTime = Date.now();
        log.debug(`Feature flag values hash generation completd for org: ${org} in ${orgFeatureFlagvaluesHashFinisedTime - startTime}ms`);

        graphQLSchema = getFilterSchemaForOrg(graphQLSchema, org, orgFeatureFlagValuesHash);
        log.debug(`added org: ${org} schema into cache in ${Date.now() - startTime}ms. Total ${orgFeatureFlagValuesHashCache.size} orgs entries and ${hashToSchemaCache.size} schema entries in the cache.`);

        //setting the cache entry to org and filter schema hash.
        orgFeatureFlagValuesHashCache.set(org, orgFeatureFlagValuesHash);
        return graphQLSchema;
    } else {
        let orgFeatureFlagValuesHash = orgFeatureFlagValuesHashCache.get(org);
        return getFilterSchemaForOrg(graphQLSchema, org, orgFeatureFlagValuesHash)
    }
}

function subscribeFeatureFlagEvents(ffNames = []) {
    for (const ffName of ffNames) {
        if (schemaFeatureFlagsSet.has(ffName)) {
            log.debug(`CSP FF change Event listner already registered for ${ffName}`);
        } else {
            cspFFClient.on(FFEvents.POLLING_FLAG_UPDATED + `:${ffName}`, () => {
                if (schemaFeatureFlagsSet.has(ffName)) {
                    log.debug(`${ffName} updated`);
                    clearCache();
                }
            });
            cspFFClient.on(FFEvents.POLLING_FLAG_DELETED + `:${ffName}`, () => {
                if (schemaFeatureFlagsSet.has(ffName)) {
                    log.debug(`${ffName} deleted`);
                    clearCache();
                }
            });
            cspFFClient.on(FFEvents.POLLING_FLAG_CREATED + `:${ffName}`, () => {
                if (schemaFeatureFlagsSet.has(ffName)) {
                    log.debug(`${ffName} created`);
                    clearCache();
                }
            });
            log.debug(`CSP FF change Event listner registered for ${ffName}`);
        }
    }

    //remove the Listeners for flags removed from the previous schema
    for (ffName of schemaFeatureFlagsSet) {
        if (!ffNames.has(ffName)) {
            cspFFClient.removeAllListeners(FFEvents.POLLING_FLAG_UPDATED + `:${ffName}`);
            cspFFClient.removeAllListeners(FFEvents.POLLING_FLAG_DELETED + `:${ffName}`);
            cspFFClient.removeAllListeners(FFEvents.POLLING_FLAG_CREATED + `:${ffName}`);
        }
    }
}

function computeOrgFeatureFlagValuesHash(org) {
    let orgFFValues = "|";
    for (const ff of schemaFeatureFlagsSet) {
        orgFFValues = orgFFValues + `${ff}:${evaluateFF(ff)}|`;
    }
    log.debug(`feature flag evaluated String for org: ${org} : ${orgFFValues}`);
    return objectHash(orgFFValues, {algorithm: "RSA-SHA256"});
}

function getFilterSchemaForOrg(graphQLSchema, org, orgFeatureFlagValuesHash) {
    if (hashToSchemaCache.has(orgFeatureFlagValuesHash)) {
        graphQLSchema = hashToSchemaCache.get(orgFeatureFlagValuesHash)
    } else {
        graphQLSchema = hashToSchemaCache.get(PUBLIC_SCHEMA_WITHOUT_FF_FILTER);

        let featureFlagDirectiveDirectiveFilterStartTime = Date.now();
        graphQLSchema = featureFlagDirectiveTransformer(graphQLSchema);
        let featureFlagDirectiveDirectiveFilterFinishedTime = Date.now();
        log.debug(`filtered the feature flag directive elements in the schema for org: ${org} in ${featureFlagDirectiveDirectiveFilterFinishedTime - featureFlagDirectiveDirectiveFilterStartTime}ms`);

        graphQLSchema = pruneSchema(graphQLSchema);
        let pruneSchemaFilterFinishedTime = Date.now();
        log.debug(`pruned schema for org: ${org} in ${pruneSchemaFilterFinishedTime - featureFlagDirectiveDirectiveFilterFinishedTime}ms`);
        //setting the schemaHash and filered schema into the schemaCache,
        // it required to avoid the duplicate entries for orgs which have same schema and most of the orgs has same schema
        // so it reduces the memory usage of cache
        hashToSchemaCache.set(orgFeatureFlagValuesHash, graphQLSchema);
    }
    return graphQLSchema;
}

//Filter schema elements starting with _
function filterUnderScoreElementsWithOutPruning(graphQLSchema) {
    return filterSchema({
        schema: graphQLSchema,
        rootFieldFilter: (type, fieldName) => !fieldName.startsWith('_'),
        fieldFilter: (type, fieldName) => !fieldName.startsWith('_'),
        argumentFilter: (type, fieldName, argName) => !(fieldName.startsWith('_') || argName.startsWith("_"))
    });
}

//Private Schema will not filter based on Feature flag but only on Feature Stability.
function createPrivateGraphQLSchema(graphQLSchema) {
    //checking the schema cache directly, because we are adding the PUBLIC and PRIVATE schema directly into schema cache
    // due to more frequent usage of these two schema.
    if (!hashToSchemaCache.has(PRIVATE)) {
        graphQLSchema = featureStabilityDirectiveTransformer(graphQLSchema);
        graphQLSchema = pruneSchema(graphQLSchema);

        //adding the PRIVATE schema directly into schema cache due to more frequent usage of this schema for service requests.
        hashToSchemaCache.set(PRIVATE, graphQLSchema);
        log.debug(`private schema added into cache. Total ${hashToSchemaCache.size} entries in the cache`);
    }
    return hashToSchemaCache.get(PRIVATE);
}

function clearCache() {
    log.debug(`clearing ${orgFeatureFlagValuesHashCache.size} schema entries from cache`);
    orgFeatureFlagValuesHashCache.clear();
    log.debug(`clearing ${hashToSchemaCache.size} org entries from cache`);
    return hashToSchemaCache.clear();
}

module.exports.createFilteredPublicGraphQLSchema = createFilteredPublicGraphQLSchema;
module.exports.createPrivateGraphQLSchema = createPrivateGraphQLSchema;
module.exports.clearCache = clearCache;
