/*
 * Copyright (c) 2022 VMware, Inc. All Rights Reserved.
 */

const {gatewayConfig} = require('./config');
const {buildSchema} = require('graphql');
const {printSchemaWithDirectives} = require("@graphql-tools/utils");
const makeRemoteExecutor = require('./make_remote_executor');
const {log} = require("./logger");
const {reqCtxt, runInContext} = require("./request_context");
const {GraphQLSchema} = require("graphql/type/schema");
const {clearCache, createFilteredPublicGraphQLSchema} = require('./schema_filter_service');
const objectHash = require('object-hash');
const config = gatewayConfig();

module.exports = class SchemaLoader {
    constructor({stitchSchemas, endpoints, loadSchemaFromFS}) {
        this.stitchSchemas = stitchSchemas;
        this.endpoints = endpoints;
        this.loadedEndpoints = new Map();
        this.schema = null;
        this.schemaHash = null;
        this.loadSchemaFromFS = loadSchemaFromFS;
    }

    async reload() {
        return await runInContext(async () => {
            reqCtxt.set("traceId", "RELOAD");
            return await this.reloadEndpoints();
        })
    }

    async reloadEndpoints() {
        let loadedEndpoints = await Promise.all(this.endpoints.map(async (endpoint) => {
            try {
                const schema = await loadSchemaForEndpoint(endpoint, this.loadSchemaFromFS);
                log.debug(`Reloaded schema for endpoint ${endpoint.url} ${endpoint.name}`);
                return {
                    endpoint,
                    schema: schema,
                };
            } catch (err) {
                log.debug(`Cannot load schema ${endpoint.url} ${endpoint.name}: ${err.message}`);
                return null;
            }
        }));

        loadedEndpoints = loadedEndpoints.filter(Boolean);
        //This ensures that if an endpoint was loaded it will continue to show in the stitched schemas. If a call fails
        //loadedEndpoints has a null value for that endpoint and it will not be loaded in the this.loadedEndpoints map.
        //so the system continues to use older version of the schema. We have the below issues to fix this:
        //TODO: ENS-4417 - Is to handle stitched schema validations.
        //TODO: ENS-2935 - Is to handle static sub schema loading instead of refreshing from endpoints.
        for (const loadedEndpoint of loadedEndpoints) {
            this.loadedEndpoints.set(loadedEndpoint.endpoint, loadedEndpoint.schema);
        }
        let schemaStichStartingTime = Date.now();
        let stitchedSchema = this.stitchSchemas(this.loadedEndpoints);
        // sort the types and directives so that the SDL is repeatable no matter which order services arrived
        let schemaConfig = stitchedSchema.toConfig();
        log.debug(`schema stitch completed in ${Date.now() - schemaStichStartingTime}ms`);

        let schemaSortStartTime = Date.now();
        schemaConfig.types.sort()
        schemaConfig.directives.sort()
        log.debug(`schema types sort completed in  ${Date.now() - schemaSortStartTime}ms`);

        this.schema = new GraphQLSchema(schemaConfig)

        let printSchemaStartTime = Date.now();
        const printedSchema = printSchemaWithDirectives(this.schema);
        log.debug(`print schema completed in ${Date.now() - printSchemaStartTime}ms`);

        let schemaHashStartTime = Date.now();
        let currentSchemaHash = objectHash(printedSchema, {algorithm: "RSA-SHA256"})
        log.debug(`schema Hash generated in ${Date.now() - schemaHashStartTime}ms`);

        if (currentSchemaHash != this.schemaHash) {
            log.info(`schema got changed. previous schema hash ${this.schemaHash} and current schema hash is ${currentSchemaHash}`);
            // clear the schema cache to add the latest schema to cache.
            clearCache();
            createFilteredPublicGraphQLSchema(this.schema);
        }
        this.schemaHash = currentSchemaHash;
        log.verbose(`gateway reload ${new Date()}, endpoints: ${this.loadedEndpoints.length}`);
        return this.schema;
    }

    autoRefresh(interval = config.schemaReloadInSec * 1000) { // 10min interval
        this.stopAutoRefresh();
        this.intervalId = setTimeout(async () => {
            let schemaReloadStartTime = Date.now();
            log.debug(`schema refresh statreted `);
            await this.reload();
            this.intervalId = null;
            log.debug(`schema refresh completed in ${Date.now() - schemaReloadStartTime}ms`);
            this.autoRefresh(interval);
        }, interval);
    }

    stopAutoRefresh() {
        if (this.intervalId != null) {
            clearInterval(this.intervalId);
            this.intervalId = null;
            log.debug(`reload timer cleared for ${this.intervalId}`);
        }
    }
};

// Custom fetcher that queries a remote schema for an "_sdl" field.
// This is NOT a standard GraphQL convention – it's just a simple way
// for a remote API to provide its own schema, complete with custom directives.
async function fetchRemoteSDL(executor, context) {
    const {data} = await executor({document: '{ _sdl }', context});
    return data._sdl;
}

async function loadSchemaForEndpoint(endpoint, loadSchemaFromFS) {
    let schemaText;
    if (loadSchemaFromFS) {
        const {data: {subSchemas}} = require("../schemas/remote/subSchemas.json");
        const subSchemaDetails = subSchemas.find((s) => s.name === endpoint.name);
        if (subSchemaDetails) {
            schemaText = subSchemaDetails.schema;
            log.debug(`Loaded schema for endpoint ${endpoint.name} from schema file`);
        } else {
            throw new Error(`Failed to load schema for ${endpoint.name} from schema file`);
        }
    } else {
        const fetcher = makeRemoteExecutor(endpoint.url, null, {timeout: config.schemaLoadTimeoutSec * 1000});
        schemaText = await fetchRemoteSDL(fetcher, {});
    }
    return buildSchema(schemaText);
}