const opentelemetry = require("@opentelemetry/api");
const {JaegerPropagator} = require('@opentelemetry/propagator-jaeger');
const {W3CTraceContextPropagator, CompositePropagator} = require('@opentelemetry/core');
const {NodeTracerProvider} = require("@opentelemetry/sdk-trace-node");
const {ConsoleSpanExporter, BatchSpanProcessor} = require("@opentelemetry/sdk-trace-base");
const {Resource} = require("@opentelemetry/resources");
const {SemanticResourceAttributes} = require("@opentelemetry/semantic-conventions");

const {registerInstrumentations} = require("@opentelemetry/instrumentation");
const {HttpInstrumentation} = require("@opentelemetry/instrumentation-http");
const {FetchInstrumentation} = require("@opentelemetry/instrumentation-fetch");
const {WinstonInstrumentation} = require('@opentelemetry/instrumentation-winston');

const {OTLPTraceExporter} = require('@opentelemetry/exporter-trace-otlp-grpc');
const {gatewayConfig} = require('./config');

// load gateway config
const config = gatewayConfig();

// register instrumentation libraries
registerInstrumentations({
    instrumentations: [
        new WinstonInstrumentation({
            // Optional hook to insert additional context to log metadata.
            // Called after trace context is injected to metadata.
            enabled: true,
            logHook: (span, record) => {
                record['resource.service.name'] = provider.resource.attributes['service.name'];
                record['parent_span_id'] = span['parentSpanId'];
            }
        }),
        new HttpInstrumentation({
            applyCustomAttributesOnSpan: (span) => {
                // some of the resonponses having the http.status_text is empty and wave front not allowed empty value in attribute
                var httpStatusText = span.attributes["http.status_text"];
                if (httpStatusText !== undefined && httpStatusText == '') {
                    span.setAttribute("http.status_text", `${span.attributes["http.status_code"]}`);
                }

                // wave front allowed max 128 length in attribute values
                var httpUrl = span.attributes["http.url"];
                if (httpUrl !== undefined && httpUrl.length > 128) {
                    var trimmedUrl = httpUrl.substring(0, 125) + "...";
                    span.setAttribute("http.url", trimmedUrl);
                }
            }
        }),
        new FetchInstrumentation()
    ],
});

const resource = Resource.default().merge(
    new Resource({
        "cluster": config.tracing.cluster,
        "application": config.tracing.applicationName,
        [SemanticResourceAttributes.SERVICE_NAME]: config.tracing.serviceName,
    })
);

const provider = new NodeTracerProvider({
    resource: resource
});

if (config.tracing.enabled) {
    const otlpOptions = {
        url: `http://${config.tracing.uri}:${config.tracing.tracingPort}`,
    };
    const otlpExporter = new OTLPTraceExporter(otlpOptions);
    const otlpProcessor = new BatchSpanProcessor(otlpExporter);
    provider.addSpanProcessor(otlpProcessor);
}
provider.register({
    //To support multiple propagations
    propagator: new CompositePropagator({
        propagators: [
            new W3CTraceContextPropagator(),
            new JaegerPropagator()
        ]
    })
});
module.exports.provider = provider;

const {log} = require('./logger');
log.info(`Gateway config is: ${JSON.stringify(config, (k, v) => (k === "cspFFSSDKKey" || k === "token") ? "xxxxxxxxxxxxxxxxxxx" : v, 2)}`)